/*--------------------------------------------------------------------
|                                                                     |
|    Convoy                                                           |
|                                                                     |
 -------------------------------------------------------------------- 
|
| Author : GeekToo
| Version: 1.2
|
|
 --------------------------------------------------------------------*/

import("queue.binary_heap", "Queue", 1);

/*--------------------------------------------------------------------
|                                                                     |
|    Tile                                                             |
|                                                                     |
 --------------------------------------------------------------------*/
class Tile
{
	constructor() {
		
	}
	function GetAdjacentTiles(tile);
	function IsRoadBuildable(tile);  // used for valuator
}	

function Tile::GetAdjacentTiles(tile)
{
	local adjTiles = AITileList();
	
	adjTiles.AddTile(tile - AIMap.GetTileIndex(1,0));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(0,1));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(-1,0));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(0,-1));
	
	return adjTiles;
}

function Tile::IsRoadBuildable(tile)
{
	if (AITile.IsBuildable(tile) || AIRoad.IsRoadTile(tile)) return true;
	return false;
}


/*--------------------------------------------------------------------
|                                                                     |
|    Node                                                             |
|                                                                     |
 --------------------------------------------------------------------*/
class Node
{
	tile_index = 0;
	predec     = 0;
	f = 0;
	g = 0;
	h = 0;

	constructor() {
		tile_index = 0;
		predec     = 0;
		f = 0;
		g = 0;
		h = 0;
	}
}

/*--------------------------------------------------------------------
|                                                                     |
|    NodeList                                                         |
|                                                                     |
 --------------------------------------------------------------------*/
class NodeList
{
	list = null;
	constructor() {
		list = {};
	}
	function Append(node);
	function Exists(a);
}

function NodeList::Append(node)
{
	list.rawset(node.tile_index,node);
}


function NodeList::Exists(a)
{
	local node = -1;
	if (list.rawin(a)){
		node = list.rawget(a);
	}
	return node;
}


/*--------------------------------------------------------------------
|                                                                     |
|   BinHeap                                                           |
|                                                                     |
 --------------------------------------------------------------------*/
class BinHeap
{
	list = null;

	constructor() {
		list = [];
	}
	function Insert(node);
	function PopMin();
	function BubbleDown();
	function Exists(a);
	function Dump();
}

/*------------------------------------------------------------------*/
function BinHeap::Insert(node)
{
//	AILog.Info("Heap: insert: node"+node.f );
	list.append(node);
	local hole = list.len();
	for( ; hole > 1 && node.f < list[ (hole / 2)-1 ].f; hole /= 2 )
		list[ hole - 1 ] = list[ (hole / 2)-1 ];
	list[ hole - 1 ] = node;
//	 Dump();
}

/*------------------------------------------------------------------*/
function BinHeap::Pop()
{	
	local min_node = null;
	if (list.len()) {
		min_node = list[ 0 ];
		list[0] = list[list.len()-1];
		BubbleDown();
		list.pop();
	}
//	Dump();
	return min_node;
}

/*------------------------------------------------------------------*/
function BinHeap::BubbleDown()
{
	local hole = 1;
   local tmp = list[ 0 ];
	
	while( hole * 2 <= list.len() ) {
		local child = hole * 2;
		if( (child != list.len() ) &&  list[ child ].f < list[ child - 1 ].f )
			child++;
		if( list[ child - 1 ].f < tmp.f )
			list[ hole - 1 ] = list[ child - 1 ];
		else
			break;
		hole = child;	
	}
	list[ hole - 1 ] = tmp;
}

/*------------------------------------------------------------------*/
function BinHeap::Exists(a)
{
	local i=0;
	foreach (list_node in list) {
		if (list_node.tile_index == a){
			return i;
		}
		i++;
	}
	return -1;
}

/*------------------------------------------------------------------*/
/** Just for debugging */
function BinHeap::Dump()
{
	local i=0;
	foreach (list_node in list) {
		AILog.Info("Heap: " + i + " " + list_node.f );
		i++;
	}
	return -1;
}


/*--------------------------------------------------------------------
|                                                                     |
|   PathFinder                                                        |
|                                                                     |
 --------------------------------------------------------------------*/
class PathFinder
{
	constructor() {
	}
	function BuildRoad(tile1, tile2);
}

/*------------------------------------------------------------------*/
function PathFinder::BuildRoad(start, target)
{
	local retry_find   = false;	
	local plan_retries = 0;

	if(start == null || target == null )	{
		AILog.Info("PathFinder: BuildRoad: null tiles");
		return false;
	}

	if( !AIMap.IsValidTile(start) || !AIMap.IsValidTile(target)) {
		return false;
	}	

	do {
		retry_find = false;
		local curr_node = Node();
		local open_list = BinHeap();
   	local closed_list = NodeList();
		local curr_tile = start;
	
		curr_node.tile_index = curr_tile;
		curr_node.g = 0; 
		curr_node.h =  10 * AITile.GetDistanceManhattanToTile(start, target); // shortest distance to target 
		curr_node.f = curr_node.h; 
		curr_node.predec = 0;

		local prev_tile = curr_node.predec;
		if (AIRoad.IsRoadTile(start)) {
			local neighbours = Tile.GetAdjacentTiles(curr_tile);
			neighbours.Valuate(AIRoad.IsRoadTile);
			neighbours.KeepValue(1);
			if (neighbours.Count()){
				neighbours.Valuate(AIRoad.AreRoadTilesConnected, curr_tile);
				neighbours.KeepValue(1);
				if (neighbours.Count()) {
					neighbours.Valuate(AITile.GetDistanceManhattanToTile, target);
					neighbours.KeepTop(1);
					prev_tile = neighbours.Begin();
				}
			}
		}
	
		local k=0;
		while ( ( curr_tile != target ) && curr_node) {	
//			AISign.BuildSign(curr_tile,"pf");
			k++;
			if ( k % 100 == 0) 
			AIController.Sleep(1);

			local neighbour_tiles = Tile.GetAdjacentTiles(curr_tile);
			
			neighbour_tiles.Valuate(Tile.IsRoadBuildable);
			neighbour_tiles.KeepValue(1);

			for ( local tile = neighbour_tiles.Begin(); neighbour_tiles.HasNext(); tile = neighbour_tiles.Next()){
				local node = Node();
				node.tile_index = tile;
				node.g = curr_node.g + 10; // default cost
				node.h =  10 * AITile.GetDistanceManhattanToTile(tile, target); // shortest distance to target 
				if (AIRoad.IsRoadTile(tile)){
					node.g -= 2;
				}
				if (AITile.GetSlope(tile)){
					node.g += 3;
				}
				node.predec = curr_node.tile_index;
				if (closed_list.Exists(tile) == -1){
					local cmd_ok = true;
					if (open_list.Exists(tile) == -1){
						if ( AITile.GetSlope(curr_tile)){
							if (prev_tile && ((curr_tile - prev_tile) != (tile - curr_tile))){
								cmd_ok = false;
							}
						}
						if ( AITile.GetSlope(tile) && AIRoad.IsRoadTile(tile)){
							local neighbour_road = Tile.GetAdjacentTiles(tile)
							neighbour_road.RemoveTile(curr_tile);
							neighbour_road.RemoveTile(tile + tile - curr_tile);
							neighbour_road.Valuate(AIRoad.IsRoadTile);
							neighbour_road.KeepValue(1);
							if (neighbour_road.Count()){
								cmd_ok = false;
							}
						} 
						if (cmd_ok) {
							node.f = node.g + node.h;
							open_list.Insert(node); 
						}
					}
				}
			}	
			
		   closed_list.Append(curr_node);
			curr_node = open_list.Pop();
			if (curr_node) {
				curr_tile = curr_node.tile_index;
				prev_tile = curr_node.predec;
			}
			else {
				return false;		
			}
		}
	
		local route_tiles = [];
		route_tiles.append(curr_node.tile_index);
//		AILog.Info("Add tile to route"+ AIMap.GetTileX(curr_node.tile_index)+" " + AIMap.GetTileY(curr_node.tile_index));
		local tile_node = Node(); 
		tile_node = curr_node;
		local finished = false;
		do {
	   	local list_node = closed_list.Exists(tile_node.predec);
			if (list_node != -1){
				tile_node = list_node;
				route_tiles.append(tile_node.tile_index);
//				AILog.Info("Add tile to route"+ AIMap.GetTileX(tile_node.tile_index)+" " + AIMap.GetTileY(tile_node.tile_index));
				finished = (tile_node.predec == 0);
			}
			else{
				finished = true;
			}
		} while (!finished);

		local route_tile_prev = route_tiles.pop();
		local err = AIError.ERR_NONE;
		if (route_tile_prev && route_tiles.len())	{
			do {
				local route_tile = route_tiles.pop();
				local retry_build = false;
				local build_retries = 0;
				do {
					retry_build = false;
					local built_road = AIRoad.BuildRoad(route_tile_prev, route_tile);
					if (!built_road) {
						switch (err = AIError.GetLastError()) {
							/* ignore these errors */
							case AIError.ERR_NONE: 	
							case AIError.ERR_ALREADY_BUILT: 
							break;
							/* can't handle this locally, return unsuccessful */
							case AIError.ERR_PRECONDITION_FAILED:
							case AIError.ERR_NEWGRF_SUPPLIED_ERROR:
							case AIError.ERR_NOT_ENOUGH_CASH:  // todo: refine this one
							case AIError.ERR_LOCAL_AUTHORITY_REFUSES: 
								AILog.Info("Build road failed(fatal)"+ AIError.GetLastErrorString()+" "+
								AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
						  		+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
					  			AITile.GetSlope(route_tile_prev));
								return false;
							break;
							case AIError.ERR_VEHICLE_IN_THE_WAY: 
								AILog.Info("Build road failed (retry)"+ AIError.GetLastErrorString()+" "+
								AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
						  		+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
						  		AITile.GetSlope(route_tile_prev));	
								AIController.Sleep(1);
								retry_build = true;
							break;
							case AIError.ERR_AREA_NOT_CLEAR: 	
								AILog.Info("Build road failed (demolish + retry)"+ AIError.GetLastErrorString()+" "+
								AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
					  			+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
					  			AITile.GetSlope(route_tile_prev));	
								AITile.DemolishTile(route_tile);
								retry_build = true;
								break;
							case AIError.ERR_OWNED_BY_ANOTHER_COMPANY :
							case AIError.ERR_FLAT_LAND_REQUIRED: 	
							case AIError.ERR_LAND_SLOPED_WRONG: 	
							case AIError.ERR_SITE_UNSUITABLE: 	
							case AIError.ERR_TOO_CLOSE_TO_EDGE: 
								AILog.Info("Build road failed (replan)"+ AIError.GetLastErrorString()+" "+
								AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
					  			+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
						  		AITile.GetSlope(route_tile_prev));	
								retry_build = false;
								retry_find = true;
								break;
							case AIError.ERR_UNKNOWN:
							default:
								AILog.Info("Build road failed (replan)"+ AIError.GetLastErrorString()+" "+
								AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
						  		+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
					  			AITile.GetSlope(route_tile_prev));	
								break;
						}
					}
					build_retries++;
				} while (retry_build && build_retries < 30 && !retry_find);	
				route_tile_prev = route_tile;
			} while ( route_tiles.len() && !retry_find);
		} 
		plan_retries++;
	}	while (retry_find && (plan_retries < 10));
	return true;
}


/*--------------------------------------------------------------------
|                                                                     |
|    TownManager                                                      |
|                                                                     |
 --------------------------------------------------------------------*/
class TownManager
{
	constructor() {

	}
	function CreateExitRoute(busstop, town1);
	function BuildBusStop(tile);
	function FindLineBusStopLocation(town, pass_cargo_id);
}

/*------------------------------------------------------------------*/
function TownManager::CreateExitRoute(busstop, town1) 
{
	local rtl = AITileList();
				
	rtl.Clear();
	rtl.AddRectangle(busstop + AIMap.GetTileIndex(-5,-5), busstop + AIMap.GetTileIndex(5,5));
	rtl.Valuate(Tile.IsRoadBuildable);
	rtl.KeepValue(1);
	rtl.Valuate(AITile.GetSlope);
	rtl.KeepValue(0);
	
	rtl.Valuate(AITile.GetDistanceManhattanToTile, AITown.GetLocation(town1));
	rtl.KeepBottom(1);
	local town_exit0 = rtl.Begin();
//	AISign.BuildSign(town_exit0,"cer");
	local connected = PathFinder.BuildRoad(busstop, town_exit0); 
	if (!connected){
		AILog.Info("CreateExitRoute failed");
	}
	return town_exit0;

}

/*------------------------------------------------------------------*/
function TownManager::FindLineBusStopLocation(town, pass_cargo_id)
{
	local tl = AITileList();
	local aitile = null;
	local tile = AITown.GetLocation(town);
	local found = false;
	
	tl.AddRectangle(tile + AIMap.GetTileIndex(-8, -8), tile + AIMap.GetTileIndex(8, 8));

	/* remove all tiles that are already covered by a station */
	local tl2 = AITileList();

	tl2.AddRectangle(tile + AIMap.GetTileIndex(-8, -8), tile + AIMap.GetTileIndex(8, 8));
	tl2.Valuate(AIRoad.IsRoadStationTile);
	tl2.KeepValue(1);
	for (local rstl = tl2.Begin(); tl2.HasNext() ; rstl = tl2.Next()){
		 tl.RemoveRectangle(rstl + AIMap.GetTileIndex(-5, -5), rstl + AIMap.GetTileIndex(5, 5));
	}	 

	if (tl.Count()) {
		/* find all tiles that are next to a road tile */
		tl.Valuate(AIRoad.GetNeighbourRoadCount);	
		tl.KeepAboveValue(0);

		if (tl.Count()) {
			/* find all tiles that are not road */
	   	tl.Valuate(AIRoad.IsRoadTile);	
   	  	tl.KeepValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not sloped */
	   	tl.Valuate(AITile.GetSlope);	
   	   tl.KeepValue(0);
	   }
   	if (tl.Count()) {
			tl.Valuate(AITile.GetCargoAcceptance, pass_cargo_id,
						 1, 1, 
						 AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP));
			aitile = tl.Begin();
			if (tl.GetValue(aitile) >= 25){
	//			AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
  				found = AITile.IsBuildable(aitile);
				if (!found)	{
					found = AITile.DemolishTile(aitile);
				}	
	   	}
		}
	}
	if (found) {
		return aitile;	
	}	
	else {
		return null;	
	}	
}



/*------------------------------------------------------------------*/
function TownManager::BuildBusStop(tile)
{
	local adjacentTiles = AITileList();
	local found = false;
	local success = false;

	local adjacentTiles = Tile.GetAdjacentTiles(tile);

	for(local tile2 = adjacentTiles.Begin(); 
			adjacentTiles.HasNext() && !found; 
			tile2 = adjacentTiles.Next()) {
		if(AIRoad.IsRoadTile(tile2) ) {
			AIRoad.BuildRoad(tile2, tile);
			local truck = false; 
			local drive_trough = false;
			success = AIRoad.BuildRoadStation(tile, tile2, truck, drive_trough);
			found = true;
		}
	}

	if (!success) {
		AILog.Info("Build busstop failed "+ AIError.GetLastErrorString());
		AISign.BuildSign(tile,"bbs");
	}
	return success;	
}


/*--------------------------------------------------------------------
|                                                                     |
|    RoutesList                                                       |
|                                                                     |
 --------------------------------------------------------------------*/
class RoutesList
{
	routes      = null;

	constructor() {
		this.routes     = AIList();
	}
	function AddItem(town1, town2);
	function HasItem(town1, town2);
}

/*------------------------------------------------------------------*/
function RoutesList::AddItem(town1, town2)
{
	if (town2 > town1){
		this.routes.AddItem((town2 << 10) + town1, 0);
	}	
	else{
		this.routes.AddItem((town1 << 10) + town2, 0);
	}
}

/*------------------------------------------------------------------*/
function RoutesList::HasItem(town1, town2)
{
	local ret = false; 
	if (town2 > town1){
		ret = this.routes.HasItem((town2 << 10) + town1);
	}	
	else{
		ret = this.routes.HasItem((town1 << 10) + town2);
	}
	return ret;
}



/*--------------------------------------------------------------------
|                                                                     |
|    RoutePlanner                                                     |
|                                                                     |
 --------------------------------------------------------------------*/
class RoutePlanner
{
	towns_used  = null;
	routes      = null;
	town_list   = null;
	town_list2  = null;
	state       = 0;
	curr_town   = 0;
	constructor() {
		this.towns_used = AIList();
		this.routes     = RoutesList();
	   this.town_list2 = AITownList(); 
		this.town_list  = AIList();
		this.town_list.AddList(this.town_list2);
		town_list.Valuate(AITown.GetPopulation);
		curr_town   = 0;
	}
	function FindUnusedTowns();
}

/*------------------------------------------------------------------*/
function RoutePlanner::FindUnusedTowns()
{


	local town    = 0;
	local town2   = 0;
	local town_it = AITown();
	local found   = false;

	switch (state) {
		case	0:	
			AILog.Info("Find unused towns, state 0");
			if (town_list.Count()){
				for (town_it = town_list.Begin(); town_list.HasNext(); 
	  				town_it = town_list.Next()){
					this.town_list.RemoveItem(town_it);
					local tl2 = AITileList();
					/* for our first town, be sure to use a town without a station */
					tl2.AddRectangle(AITown.GetLocation(town_it) + AIMap.GetTileIndex(-10,-10),
				   					AITown.GetLocation(town_it) + AIMap.GetTileIndex(10,10));
					tl2.Valuate(AIRoad.IsRoadStationTile);
					tl2.KeepValue(1);
					if (!tl2.Count()){
						town = town_it;
						this.towns_used.AddItem(town, 0);
						found = true;
						break;
					}
				}		
				if (found) {
					town_list2.Clear();
					town_list2.AddList(town_list);
					town_list2.Valuate(AITown.GetDistanceManhattanToTile, AITown.GetLocation(town));
					town_list2.KeepBelowValue(120);
					town_list2.KeepBottom(6);
					town_list2.Valuate(AITown.GetPopulation);
					for (town_it = town_list2.Begin(); town_list2.HasNext(); 
	   					town_it = town_list2.Next()){
						town_list.RemoveItem(town_it);
						local tl2 = AITileList();
						tl2.AddRectangle(AITown.GetLocation(town_it) + AIMap.GetTileIndex(-10,-10),
							   				AITown.GetLocation(town_it) + AIMap.GetTileIndex(10,10));
						tl2.Valuate(AIRoad.IsRoadStationTile);
						tl2.KeepValue(1);
						if (!tl2.Count()){
							town2 = town_it;
							this.towns_used.AddItem(town2, 0);
							found = true;
							break;
						}
						else {
							found = false;
						}
					}
					if (!found) {
						this.towns_used.Valuate(AITown.GetDistanceManhattanToTile, AITown.GetLocation(town));
						this.towns_used.Sort(AIList.SORT_BY_VALUE, true);
						for (town_it = this.towns_used.Begin(); this.towns_used.HasNext(); 
   						town_it = this.towns_used.Next()){
							if (town_it != town){
								if (!this.routes.HasItem(town_it, town)){
									found = true;
									town2 = town_it;
									this.towns_used.AddItem(town2, 0);
									break;
								}
							}
						}	
					}
				}		
				if (found) {
					this.routes.AddItem(town, town2);
				}	
			}
			else {
				state = 1;
				town_list2.Clear();
				town_list2.AddList(this.towns_used);
				town_list2.Valuate(AITown.GetPopulation);
				curr_town = town_list2.Begin();
				return null;
			}	
		break;
		case 1:
			break; 
			AILog.Info("Find unused towns, state 1");
			found = false;
			town = curr_town;
			this.towns_used.Valuate(AITown.GetDistanceManhattanToTile, AITown.GetLocation(town));
			this.towns_used.Sort(AIList.SORT_BY_VALUE, true);
			for (town_it = this.towns_used.Begin(); this.towns_used.HasNext(); 
				town_it = this.towns_used.Next()){
				if (town_it != town){
					if (!this.routes.HasItem(town_it, town)){
						found = true;
						town2 = town_it;
						this.routes.AddItem(town, town2);
						break;
					}
				}
			}	
			if (town_list2.HasNext()) { 
				curr_town = town_list2.Next()
			} else {
				curr_town = town_list2.Begin();
			}
		break;
	}
	if (found) {
		return [ town, town2 ];
	}
	else {
		return null;
	}
}

/*--------------------------------------------------------------------
|                                                                     |
|    Line                                                             |
|                                                                     |
 --------------------------------------------------------------------*/

class Line
{
	towns    = null;
	stations = null;
	depot    = null;
	vehicles = null;
	date_last_vehicle = 0;
	failed = false;
	passenger_cargo_id = 0;
	
	constructor(cargo_id) {
		towns     = [];
		stations = [[], []];
		depot    = null;
		vehicles  = [];
		failed = false;
		date_last_vehicle = 0;
		passenger_cargo_id = cargo_id;
	}
	function AddVehicles();
	function EstimateBusesNeeded();
	function AddDepot();
	function CreateNewLine(town_pair);
}

function Line::CreateNewLine(town_pair)
{
	local connected = false;
	local town_exit = [0, 0];
	local success   = [false, false];
	local newline   = (town_pair != null);

	if (town_pair) {
		towns = town_pair;
	}
	if(towns && (towns.len() == 2))
	{
		AILog.Info("New line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]));

		local town_idx     = 0;
		local station_idx  = 0;
		local new_location = 0;
		
		foreach (town in towns){
			new_location = TownManager.FindLineBusStopLocation(towns[town_idx],
																					passenger_cargo_id);
			if (new_location){
				stations[town_idx].append (new_location);				
				success[town_idx] = TownManager.BuildBusStop(new_location);
				if (success[town_idx]) {
					AILog.Info("busstop " + town_idx + " built");
					town_exit[town_idx] =  TownManager.CreateExitRoute(AIRoad.GetRoadStationFrontTile(new_location), towns[(town_idx+1) %2 ]);
					if (!town_exit[town_idx]) {
						success[town_idx] = false;
						AILog.Info("Create exit " + town_idx + " failed");
						break;
					}					
				}
			}	
			else {
				break;
			}
			town_idx++;
		}
					
		// Build depot

		if((success[0] == false) || (success[1] == false)) {
			AILog.Info("Failed to build bus stations");
		}
		else {
			if (newline) {
				depot = AddDepot(stations[0][0]);
				if (depot) {
					AILog.Info("Depot built");
					connected = true;
				}
				else {
					AILog.Info("Failed to build depot");
				}
			}
			else {
				connected = true;
			}
		}
		
		if (connected && newline) {
			connected = PathFinder.BuildRoad(town_exit[0], town_exit[1]); 
		}
		
		// make sure the depot and station are connected
		if(connected) {
			connected = PathFinder.BuildRoad(AIRoad.GetRoadStationFrontTile(stations[0][0]), AIRoad.GetRoadDepotFrontTile(depot)) ;
		} 
		else {
			AILog.Info("Connect exits failed "+ AIError.GetLastErrorString());
		}
		// make sure the depot and town_exit are connected for a nicer route
		if(connected){
			connected = PathFinder.BuildRoad(town_exit[0], AIRoad.GetRoadDepotFrontTile(depot)) ;
		} 
		else {
			AILog.Info("Connect busstop 0 and depot failed"+ AIError.GetLastErrorString());
		}	
	}	
	else {
		AILog.Info("no 2 towns to connect");
	}
	return connected;
}

/*------------------------------------------------------------------*/
function Line::AddVehicles()
{
	if(this.stations.len() != 2) {
		return null;
	}
	local bus_model = null;
	local min_reliability = 85;
	local new_bus;
	local i;
	local n_buses = 6;

	local engine_list = AIEngineList(AIVehicle.VEHICLE_ROAD);

	local balance = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	engine_list.Valuate(AIEngine.GetPrice);
	engine_list.KeepBelowValue(balance);

	engine_list.Valuate(AIEngine.GetCargoType)
	engine_list.KeepValue(passenger_cargo_id); 

	engine_list.Valuate(AIEngine.GetCapacity)
	engine_list.KeepTop(1);

	bus_model = engine_list.Begin();
	if(bus_model == null)
	{
		AILog.Info("BuyBuses: failed to find a bus to build");
		return false;
	}

//	n_buses = EstimateBusesNeeded();	
	// Buy buses
	local buses_build = 0;
	for(i = 0; i < n_buses; ++i)
	{
		new_bus = AIVehicle.BuildVehicle(this.depot, bus_model);
		if ( new_bus && AIVehicle.IsValidVehicle(new_bus)) { 
			this.vehicles.append(new_bus);
			this.date_last_vehicle = AIDate.GetCurrentDate();
			buses_build++;
		}
	}

	local idx0 = 0;
	local idx1 = 1;	
	local town_idx = 0;	
	for (local bus_idx = this.vehicles.len() - buses_build; bus_idx < this.vehicles.len(); bus_idx++){
		local bus = this.vehicles[bus_idx]; 

		if(bus == null || !AIVehicle.IsValidVehicle(bus)) {
			AILog.Info("Vehicle[ " + bus + "] is not valid!");
		}	
		else {
			AILog.Info("AddVehicle " + idx0 + " " + idx1);
			idx0 = this.stations[town_idx].len()-1;
			AIOrder.AppendOrder(bus, this.depot, AIOrder.AIOF_SERVICE_IF_NEEDED);
			AIOrder.AppendOrder(bus, this.stations[town_idx][idx0], AIOrder.AIOF_NONE); 
			town_idx = (town_idx + 1) % 2;
			idx1 = this.stations[town_idx].len()-1;
			AIOrder.AppendOrder(bus, this.stations[town_idx][idx1], AIOrder.AIOF_NONE);
			/* don't increment town_idx here, so orders are alternating: half the buses
			 * to 1 town, the others to the other town */
			
			AIVehicle.StartStopVehicle(bus);
		}
	}	
}

/*------------------------------------------------------------------*/
function Line::EstimateBusesNeeded()
{

	if(!towns || (towns.len() != 2))
		return 0;
	
	local acceptance = AITile.GetCargoAcceptance(AITown.GetLocation(towns[0]), 
																passenger_cargo_id, 1, 1, 
																AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) + 
							AITile.GetCargoAcceptance(AITown.GetLocation(towns[1]), 
																passenger_cargo_id, 1, 1, 
							AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) ;
	local distance = AIMap.DistanceManhattan( AITown.GetLocation(towns[0]), AITown.GetLocation(towns[1]) );

	AILog.Info("EstimateBusesNeeded: distance:    " + distance + 
	 										 " acceptance:  " + acceptance);

	local num_bus = 1 + (acceptance / 8) + (distance / 20);

	AILog.Info("Buy " + num_bus + " buses");
	return num_bus;
}


/*------------------------------------------------------------------*/
function Line::AddDepot(tile)
{
	local tl     = AITileList();
	local x      = AIMap.GetTileX(tile);
	local y      = AIMap.GetTileY(tile);
	local aitile = 0;
	local found  = false;
	
	tl.AddRectangle(AIMap.GetTileIndex(x-10,y-10), AIMap.GetTileIndex(x+10,y+10));
	if (tl.Count())
	{
	   if (tl.Count()) {
	   	/* find all tiles that are next to a road tile */

	   	tl.Valuate(AIRoad.GetNeighbourRoadCount);	
   	   	tl.KeepAboveValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not road */
	   	tl.Valuate(AIRoad.IsRoadTile);	
   	  	tl.KeepValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not sloped */

	   	tl.Valuate(AITile.GetSlope);	
   	   tl.KeepValue(0);
	   }
   	if (tl.Count()) {
			tl.Valuate(AITile.GetDistanceManhattanToTile, AITown.GetLocation(towns[1]));
			tl.Sort(AITileList.SORT_BY_VALUE, true);
		
			for (aitile = tl.Begin();tl.HasNext() && !found; 
			    aitile = tl.Next()) {    
//				AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
				local adjacentTiles = Tile.GetAdjacentTiles(aitile);

				/*Loop through all adjacent tiles*/

				for(local tile2 = adjacentTiles.Begin(); 
			   	 adjacentTiles.HasNext() && !found; 
			    	 tile2 = adjacentTiles.Next()) {
			   	if(AIRoad.IsRoadTile(tile2) && !AITile.GetSlope(tile2) ) {
						if (!AIRoad.IsRoadTile(tile2+tile2-aitile)) {
							if(!AIRoad.IsRoadStationTile(aitile)) {
				   			found = AITile.IsBuildable(aitile);
								if (!found)	{
									found = AITile.DemolishTile(aitile);
								}	
								if (found) {
     								AIRoad.BuildRoad(tile2, aitile);
									AIRoad.BuildRoadDepot(aitile, tile2)
								}
							}
						}
					}
		   	}
				if (found) {
			   	break;
				}
			}
	   }
	}
	if (found) {
		return aitile;	
	}
	else {
		return null;
	}
}

/*--------------------------------------------------------------------
|                                                                     |
|    ConvoyAI                                                         |
|                                                                     |
 --------------------------------------------------------------------*/
class ConvoyAI extends AIController {
	company            = null;
	stop               = false;
	vehicle_to_depot   = {};
	route_planner      = null;
	lines              = [];
	name               = null;
	passenger_cargo_id = 0;

	constructor() {
		this.company  = AICompany();
		route_planner = RoutePlanner();
		name = null;
	}

	function Start();

	function SetCompanyName();
	function ManageLoan();
	function HasMoney(money);
}

/*------------------------------------------------------------------*/
function ConvoyAI::HasMoney(money)
{
//	AILog.Info("Has money: balance " + AICompany.GetBankBalance(AICompany.MY_COMPANY)) ;
	if (AICompany.GetBankBalance(AICompany.MY_COMPANY) + (AICompany.GetMaxLoanAmount() - this.company.GetLoanAmount()) > money) return true;
	return false;
}

/*------------------------------------------------------------------*/
function ConvoyAI::Start()
{
	local line_idx = 0;
	
	this.SetCompanyName();
	this.name = AICompany.GetCompanyName(AICompany.MY_COMPANY);

	local list = AICargoList();
	for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
		if (AICargo.HasCargoClass(i, AICargo.CC_PASSENGERS)) {
			this.passenger_cargo_id = i;
			break;
		}
	}
	local i = 0;
	while(!this.stop)
	{
		this.Sleep(1);
		if((i%200)==0)
		{
			AILog.Info("Manage loan and Vehicles");
			ManageLoan();
			ManageVehicles();
		}	
		// ... check if we can afford to build some stuff
		if((i%50)==0) {
			if(this.HasMoney(25000))
			{
				AILog.Info("Start building");		

				local line = null;
				this.company.SetLoanAmount(AICompany.GetMaxLoanAmount());
				local towns = this.route_planner.FindUnusedTowns(); 
				if(towns && (towns.len() == 2)) {
					line = Line(this.passenger_cargo_id);
					local connected = line.CreateNewLine(towns);
					if(connected) {	
						AILog.Info("New line completed");
						line.AddVehicles(); 
						lines.append(line);
					}
					else {
						AILog.Info("Failed to add new line");
					}
				}
				else {
					if (lines && lines.len()){
						line_idx = (line_idx + 1) % lines.len();
						line = lines[line_idx];
						local connected = line.CreateNewLine(null);
						if(connected) {	
							AILog.Info("Update line completed");
							line.AddVehicles(); 
						}
						else {
							AILog.Info("Failed to update line");
						}
					}
				}
				// Pay back unused money
				ManageLoan();
			}
		}
		i++;
	}	
}


/*------------------------------------------------------------------*/
function ConvoyAI::SetCompanyName()
{
	if (!AICompany.SetCompanyName("Convoy")) {
		local i = 2;
		while (!AICompany.SetCompanyName("Convoy #" + i)) {
			i++;
		}
	}
}

/*------------------------------------------------------------------*/
function ConvoyAI::ManageLoan()
{
	local balance = this.company.GetBankBalance(AICompany.MY_COMPANY);
	local loan = this.company.GetLoanAmount();
	local pay_back_balance =  2 * this.company.GetLoanInterval(); 
	local loan_interval = this.company.GetLoanInterval();
	local pay_back = 0;
	
	while(( balance - pay_back >= pay_back_balance) 
			&& (loan - pay_back > 0))	{
		pay_back += loan_interval;
	}
	if (pay_back) {
		if(!this.company.SetLoanAmount(loan - pay_back))	{
			AILog.Info(this.name + " Failed to pay back");
		}
		else {
			AILog.Info("Paid back: " + pay_back);
		}
	}	
	AILog.Info("Loan: " + this.company.GetLoanAmount());
}


/*------------------------------------------------------------------*/
function ConvoyAI::ManageVehicles()
{
	if (AIDate.GetMonth(AIDate.GetCurrentDate()) > 8) {
	
		local list = AIVehicleList();
		list.Valuate(AIVehicle.GetAge);
		list.KeepAboveValue(700);
		list.Valuate(AIVehicle.GetProfitThisYear);
   
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			local profit = list.GetValue(i);

			if (profit < 0) {
				/* Send the vehicle to depot if we didn't do so yet */
				if (!this.vehicle_to_depot.rawin(i) || this.vehicle_to_depot.rawget(i) != true) {
					AILog.Info(this.name + ": [INFO] Sending " + i + " to depot as	profit is: " + profit );
					AIVehicle.SendVehicleToDepot(i);
					this.vehicle_to_depot.rawset(i, true);
				}
			}
			/* Try to sell it over and over till it really is in the depot */
			if (this.vehicle_to_depot.rawin(i) && this.vehicle_to_depot.rawget(i) == true) {
				if (AIVehicle.SellVehicle(i)) {
					AILog.Info(this.name + ": [INFO] Selling " + i + " as it finally is in a depot.");
					/* Check if we are the last one serving those airports; else sell the airports */

					this.vehicle_to_depot.rawdelete(i);
				}
			}
		}
	}

	if (!this.HasMoney(12000)) return;

	foreach (line_it in lines) {
		local curr_station = AIStation.GetStationID(line_it.stations[0][0]);
		local veh_list_station = AIVehicleList_Station(curr_station);
		
		if (veh_list_station.Count() < 20) {
			local waiting = AIStation.GetCargoWaiting(curr_station,
																	passenger_cargo_id) +
	   			AIStation.GetCargoWaiting(AIStation.GetStationID(line_it.stations[1][0]),
																		passenger_cargo_id);
			if (waiting > 250) {
				AILog.Info(this.name + ": [INFO] extra buses needed");	
				if (AIDate.GetCurrentDate() - line_it.date_last_vehicle > 50) {
					if (line_it.vehicles[0] && line_it.depot) {
						for(local i = 0; i < (waiting / 120); ++i) {
							local new_veh = AIVehicle.CloneVehicle (line_it.depot, 
																	line_it.vehicles[0], true);			
							AIVehicle.StartStopVehicle(new_veh);
							line_it.vehicles.append(new_veh);				
							line_it.date_last_vehicle = AIDate.GetCurrentDate();
						}
					}
				}
			}
		}
	}
}


